# -*- coding: utf-8 -*-
# Copyright 2021-2022, Hojin Koh
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import json
import os
import threading
from contextlib import contextmanager
from http.client import HTTPConnection
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from random import randint
from urllib.parse import quote_plus

from .logging import logger

mCache = {}
sock = None # The IPC server for exchanging cached task information


def isAvailable():
    return 'EIKTHYR_CACHE_IP' in os.environ and 'EIKTHYR_CACHE_PORT' in os.environ

@contextmanager
def openConnection():
    try:
        h = HTTPConnection(os.environ['EIKTHYR_CACHE_IP'], int(os.environ['EIKTHYR_CACHE_PORT']))
        yield h
    finally:
        h.close()


def getObj(obj):
    with openConnection() as h:
        h.request('GET', '/{}'.format(quote_plus(repr(obj))))
        r = h.getresponse()
        if r.status == 200:
            if 'Content-Length' in r.headers:
                return json.loads(r.read(int(r.headers['Content-Length'])))
            else:
                return json.loads(r.read())
        else:
            return None

def putObj(obj, content):
    with openConnection() as h:
        data = bytes(json.dumps(content), 'ascii')
        h.request('PUT', '/{}'.format(quote_plus(repr(obj))), body=data, headers={'Content-Length': len(data)})
        h.getresponse()

def deleteObj(obj):
    with openConnection() as h:
        h.request('DELETE', '/{}'.format(quote_plus(repr(obj))))
        h.getresponse()


## === Server Part Below ===

class RequestHandler(BaseHTTPRequestHandler):
    # Suppress the log message from HTTP request
    def log_message(self, format, *args):
        return

    def do_GET(self):
        if self.path in mCache:
            self.send_response(200)
            self.send_header('Content-Type', "text/json")
            self.send_header('Content-Length', len(mCache[self.path]))
            self.end_headers()
            self.wfile.write(mCache[self.path])
        else:
            self.send_response(404)
            self.end_headers()

    def do_PUT(self):
        mCache[self.path] = self.rfile.read(int(self.headers['Content-Length']))
        self.send_response(201)
        self.end_headers()

    def do_DELETE(self):
        if self.path in mCache:
            del mCache[self.path]
        self.send_response(201)
        self.end_headers()

def startCache():
    # Start the cacche server if not already
    global sock
    if sock == None:
        sock = ThreadingHTTPServer(('127.{:d}.{:d}.{:d}'.format(randint(1,251), randint(1,251), randint(1,251)), 0), RequestHandler)
        thr = threading.Thread(target=sock.serve_forever, daemon=True)
        thr.start()
        ip, port = sock.server_address
        logger.debug("Task cache server started at {}:{}".format(ip, port))
        os.environ['EIKTHYR_CACHE_IP'] = ip
        os.environ['EIKTHYR_CACHE_PORT'] = str(port)

@atexit.register
def stopCache():
    global sock
    if sock != None:
        sock.shutdown()
        sock.server_close()
        sock = None
        logger.debug('Task cache server stopped')



# -*- coding: utf-8 -*-
# Copyright 2021-2022, Hojin Koh
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

import luigi as lg
from luigi.task import flatten

from .data import Target
from .task import Task

class WhateverParameter(lg.Parameter):
    def _warn_on_wrong_param_type(self, param_name, param_value):
        return

class PathParameter(lg.PathParameter):
    def _warn_on_wrong_param_type(self, param_name, param_value):
        Path(param_value)

    def serialize(self, x):
        pathForShow = Path(x)
        if pathForShow.is_absolute():
            if pathForShow.is_relative_to(Path.cwd()):
                pathForShow = pathForShow.relative_to(Path.cwd())
        return str(pathForShow)

class TaskParameter(lg.Parameter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _warn_on_wrong_param_type(self, param_name, param_value):
        if self.__class__ != TaskParameter:
            return
        if not isinstance(param_value, Task):
            raise ValueError("parameter {} must be a Eikthyr task, got {} instead".format(param_name, param_value))

    def serialize(self, x):
        try:
            return x.output().pathRel
        except:
            try:
                return ';'.join(sorted([y.pathRel for y in flatten(x.output())]))
            except:
                return super().serialize(x)

class TaskListParameter(lg.Parameter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _warn_on_wrong_param_type(self, param_name, param_value):
        if self.__class__ != TaskListParameter:
            return
        for t in param_value:
            if not isinstance(t, lg.Task):
                raise ValueError("parameter {} must be a list of Eikthyr task, got {} instead".format(param_name, param_value))

    def serialize(self, xs):
        try:
            return ';'.join(sorted([x.output().pathRel for x in xs]))
        except:
            try:
                return ';'.join(sorted([y.pathRel for x in xs for y in flatten(x.output())]))
            except:
                return super().serialize(xs)

class TargetParameter(lg.Parameter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _warn_on_wrong_param_type(self, param_name, param_value):
        if self.__class__ != TargetParameter:
            return
        if not isinstance(param_value, Target):
            raise ValueError("parameter {} must be a Eikthyr target, got {} instead".format(param_name, param_value))

    def serialize(self, x):
        try:
            return x.pathRel
        except:
            return super().serialize(x)



# -*- coding: utf-8 -*-

from dotenv import load_dotenv as _load_dotenv
_load_dotenv()

# Deal with luigi's annoying deprecation warning
import warnings as _warnings
with _warnings.catch_warnings():
    _warnings.simplefilter("ignore")
    import luigi as lg

from . import data
from .data import Target, ConfigData

from . import cmd
from .cmd import chdir, mkcd, withEnv, cmdfmt

from . import task
from .task import Task, STask

from . import envcheck
from .envcheck import EnvCheck

from . import param
from .param import WhateverParameter, TaskParameter, TaskListParameter, TargetParameter, PathParameter
from luigi import Parameter, BoolParameter, IntParameter, FloatParameter, ListParameter, DictParameter

from . import specialtask
from .specialtask import InputTask, TargetWrapperTask, StampTask

from . import logging
from .logging import logger

from . import run
from .run import run

from luigi import Config

from plumbum import local
cmd = local.cmd
